/* nylon - (socks{4,5}) proxy
 * http://mesh.eecs.umich.edu/projects/nylon
 * marius aamodt eriksen (marius@umich.edu) */

#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <errno.h>
#include "mnet.h"
#include <sys/ioctl.h>
#include <net/if.h>

#if defined(__sun__)
#include <sys/sockio.h>
#endif /* __sun__ */

#define DEFAULT_MALLOC 8   /* userids usually aren't that large */
#define IFMAX IFNAMSIZ
#define S4_MAX_USERID 256

void usage(char**);
void fatal_error(char*);
void die(char*);
void printd(char*);
void wait_for_connections(int, char[IFMAX]); 
void negotiate_connection(int);
void negotiate4(int);
void negotiate5(int);
void establish_connection(int, int);
void sigchld(int);

typedef struct {
	unsigned char  vn,
		           cd;
	unsigned short dport;
	unsigned long  dip;
	unsigned char  terminator;
} socks4_hdr;

typedef struct {
	unsigned char	  vn,
		              cd,
		             rsv,
		            atyp;
	unsigned long  daddr;
	unsigned short dport;
} socks5_req;

/* version reply */
typedef struct {
	unsigned char ver,
		          res;
} socks5_vrepl;  

int verbose, no_connections;
char cvsid[] = "$Id: nylon.c,v 1.5 2001/06/08 04:53:51 beriksen Exp $";

/* main() - read in command line options and act upon them, then
 * fork-and-exit for ``daemon like'' behavior */

int main (int argc, char **argv)
{
	int opt, port = 1080, bgproc;
	char inf[IFMAX];

	verbose        = 0;
	no_connections = 1;
	inf[0] = '\0';

	while ((opt = getopt(argc, argv, "hvVp:n:i:")) != -1) 
		switch (opt) {
		case 'h':
			usage(argv);
		case 'v':
			verbose = 1;
			break;
		case 'V':
			printf("nylon version " VERSION ":\n%s\n", cvsid);
			exit(0);
		case 'p':
			port = atoi(optarg);
			break;
		case 'n':
			no_connections = atoi(optarg);
			break;
		case 'i':
			strlcpy(inf, optarg, IFMAX);
		default:
			break;
		}

	signal(SIGCHLD, sigchld);    /* get 'em zombies away! */

	bgproc = fork();

	if (bgproc == -1) 
		fatal_error("fork() failed");
	if (bgproc != 0) {
		printf("nylon launched with pid %i\n", bgproc); 
		exit(0);
	}
	/* we are now the child */
	wait_for_connections(port, inf);
	exit(0);
}

void wait_for_connections(int portno, char inf[IFMAX])
{
	struct sockaddr_in serv_addr, cli_addr, *bogusin;
	int sfd, nsfd, nproc, bogusfd;
	socklen_t cli_len = sizeof(struct sockaddr);
	struct ifreq ifr;
	char ifip[128];
	
	bzero(&serv_addr, sizeof(serv_addr));
	bzero(&cli_addr , sizeof(cli_addr));

	if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("socket() failed");

	serv_addr.sin_family      = AF_INET;
	serv_addr.sin_port        = htons(portno);
	if (inf[0]) {
		strlcpy(ifr.ifr_name, inf, IFMAX);
		bogusfd = socket(AF_INET, SOCK_STREAM, 0);
		ioctl(bogusfd, SIOCGIFADDR, &ifr);
		bogusin = (struct sockaddr_in *) &ifr.ifr_addr;
		memcpy(&serv_addr.sin_addr.s_addr, &bogusin->sin_addr.s_addr,
			   sizeof(serv_addr.sin_addr.s_addr));
		inet_ntop(AF_INET, &bogusin->sin_addr, ifip, sizeof(ifip));
		if (verbose)
			printf("binding to address %s\n", ifip);
	}
	else
		serv_addr.sin_addr.s_addr = INADDR_ANY;
		
	if (bind(sfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		fatal_error("bind() failed");

	if (listen(sfd, no_connections) == -1)
		fatal_error("listen() failed");

	if (verbose)
		printf("starting to listen...\n");
	for (;;) {
		if ((nsfd = accept(sfd, (struct sockaddr*) &cli_addr, &cli_len)) == -1 )
			fatal_error("accept() failed");

		nproc = fork();
		if (nproc == -1)
			fatal_error("fork() failed");
		if (nproc == 0) {
			if (verbose)
				printf("connect! starting negotiations\n");
			negotiate_connection(nsfd);
			exit(0);
		}
		close(nsfd);
	}
}

/* multiplex depending on version number */
void negotiate_connection(int sfd)
{
	unsigned char vn;

	if ( recvn(sfd, &vn, 1, 0) == -1 )
		fatal_error("error receiving version number\n");

	if (vn == 4)               /* SOCKS4 */
		negotiate4(sfd);
	else if (vn == 5)          /* SOCKS5 */
		negotiate5(sfd);
	else
		die("No support for requested protocol version\n");
}

void negotiate4(int sfd)
{
	unsigned char vn, cd, *userid, *methods, method;
	unsigned short dport;
	unsigned long dip;
	unsigned i = 0, j = 1, nmethods;
	struct sockaddr_in target_addr;
	int target_sfd;
	socks4_hdr hdr4;

	if (verbose)
		printf("entering SOCKS4 specific negotiation");

	userid = (unsigned char*)malloc(DEFAULT_MALLOC);

	hdr4.vn = 4;
	if ( recvn(sfd, &hdr4.cd, 8, 0) == -1 )
		fatal_error("error receiving header\n");

	if (verbose)
		printf("getting user id\n");

	while (recvn(sfd, userid+i, 1, 0) != -1)
		if ((userid[i]) == 0) {
			userid[i] = '\0';
			hdr4.terminator = 0;
			break;
		}
		else {
			if ((++i)%DEFAULT_MALLOC == 0)
				userid = (unsigned char*)realloc(userid, (++j)*DEFAULT_MALLOC);
			if ((j*DEFAULT_MALLOC) > S4_MAX_USERID)
				fatal_error("header too long, bailing out");
		}

	
	if (hdr4.cd != 1)
		die("SOCKS4: unsupported command!\n");

	if ( (target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("could not create socket");

	bzero(&target_addr, sizeof(target_addr));
	target_addr.sin_family      = AF_INET;
	target_addr.sin_port        = hdr4.dport;   /* already in network order */
	target_addr.sin_addr.s_addr = hdr4.dip;     /* ditto                    */

	if (verbose)
		printf("trying to connec to target host...");

	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) 
		hdr4.cd = 91;
	else
		hdr4.cd = 90;

	if ( sendn(sfd, &hdr4, 9, 0) == -1 )
		fatal_error("error replying to initiating host");

	if (cd == 91)
		fatal_error("error connecting to target host");
	
	if (verbose)
		printf("connected!\n");

	establish_connection(sfd, target_sfd);
	
	return;
}


void negotiate5(int sfd)
{
	unsigned i = 0, j = 1, reqlen = sizeof(socks5_req);
	struct sockaddr_in target_addr;
	int target_sfd;
	socks5_req req5;
	socks5_vrepl rep5;
	unsigned char nmethods, *methods, method;
	
	if (verbose)
		printf("initiating SOCKS5 connection\n");

	method  = 0;
	req5.vn = 5;

	if ( recvn(sfd, &nmethods, 1, 0) == -1 )
		fatal_error("error in receiving header");
	if (verbose)
		printf("got methodn == %i\n", nmethods);
	methods = (unsigned char*)malloc(nmethods);
	if ( recvn(sfd, methods, nmethods, 0) == -1 )
		fatal_error("error in receiving header");
	if (verbose)
		printf("got methods\n");
	/* we don't support any authentication methods yet, so simply
	 * ignore it and send reply with no authentication required. */
	if (verbose) {
		printf("they support method: %i\n", *methods);
		printf("sending reply: %i | %i\n", req5.vn, method);
	}

	free(methods);

	rep5.ver = 5;
	rep5.res = 0;

	if ( sendn(sfd, &rep5, 2, 0) == -1)
		fatal_error("error in sending reply");

	if (verbose)
		printf("sent no auth needed\n");

	if ( recvn(sfd, &req5, 4, 0) == -1 )
		fatal_error("Error receiving SOCKS5 request");
	if (req5.vn != 5)
		die("Error in header");
	if (req5.cd != 1)
		die("Unsupported SOCKS5 request");

	if (verbose)
		printf("request == %i\n", req5.cd);

	bzero(&target_addr, sizeof(target_addr));

	if (req5.atyp == 1) {   /* IPV4 Addr */
		if (verbose)
			printf("they want ipv4 addr\n");
		if (recvn(sfd, &req5.daddr, 4, 0) == -1)
			fatal_error("Error receiving request");
		target_addr.sin_family      = AF_INET;
		target_addr.sin_addr.s_addr = req5.daddr;

    } else if (req5.atyp == 3) {  /* domain name */
		unsigned char len;
		char *hostname;
		struct hostent *hent;
		if (recvn(sfd, &len, 1, 0) == -1)
			fatal_error("error receiving request");
		if (verbose)
			printf("domain name %i characters long\n", len);
		hostname = (char*)malloc(len);
		if (recvn(sfd, &hostname, len, 0) == -1)
			fatal_error("error receiving hostname");
		hostname[len] = '\0';
		hent = gethostbyname(hostname);
		target_addr.sin_family      = AF_INET;
		target_addr.sin_addr.s_addr = htonl(*hent->h_addr);
		free(hostname);
	} else 
		die("Unsupported address type");

	if (recvn(sfd, &req5.dport, 2, 0) == -1)
		fatal_error("Error receving request");
	target_addr.sin_port = req5.dport;

	if (verbose) {
		printf("target host:\n");
		printf("ip: %s\n", inet_ntoa(target_addr.sin_addr));
		printf("p : %i\n", ntohs(req5.dport));
	}

	if ( (target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		fatal_error("could not create socket");

	if (verbose)
		printf("making connection...\n");

	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) {
		if (verbose)
			printf("unsuccessful connect!\n");
		req5.cd = 1;
	} else {
		if (verbose)
			printf("successful connect!\n");
		req5.cd = 0;
	}

	if (verbose)
		printf("made connection...\n");
	
	if ( sendn(sfd, &req5, 10, 0) == -1 )
		fatal_error("error replying to initiating host");

	if (req5.cd == 1)
		fatal_error("error connecting to target host");
	
	establish_connection(sfd, target_sfd);
	
	return;
}


void establish_connection(int client, int target)
{
	fd_set rfds;
	int nfds;
	struct timeval tmout;
	unsigned char buf[1024];
	unsigned bcount;


	nfds = (client > target?client:target)+1;
	
	for (;;) {
		tmout.tv_usec = 0; tmout.tv_sec = 60;
		FD_ZERO(&rfds);
		FD_SET(client, &rfds); FD_SET(target, &rfds);
		select(nfds, &rfds, NULL, NULL, &tmout);
		bzero(buf, 1024);
		if (FD_ISSET(client, &rfds)) {
			if ( (bcount = recv(client, buf, 1024, 0)) == -1 )
				fatal_error("error in recv()");
			if (bcount == 0)
				if (verbose)
					die("client terminated connection");
				else
					exit(0);
			if (verbose)
				printf("%i bytes client->target\n", bcount);
			if ( send(target, buf, bcount, 0) == -1 )
				fatal_error("error in send()");
		} else if (FD_ISSET(target, &rfds)) {
			if ( (bcount = recv(target, buf, 1024, 0)) == -1 )
				fatal_error("error in recv()");
			if (bcount == 0)
				if (verbose)
					die("server terminated connection");
				else
					exit(0);
			if (verbose)
				printf("%i bytes target->client\n", bcount);
			if ( send(client, buf, bcount, 0) == -1 )
				fatal_error("error in send()");			
		} else
			fatal_error("select() failed");
	}
}

void sigchld (int sign)
{
	pid_t pid;
	int   stat;

	pid = wait(&stat);
	return;
}

void usage (char **argv)
{
	printf("mproxy version " VERSION " usage:\n"
		   "%s -hvV [-p portno] [-i if]\n"
		   "   h - help (this)\n"
		   "   v - verbose\n"
		   "   V - version\n"
		   "   i - bind to specified interface\n"		   
		   "   p - port number\n", *argv);
	exit(0);
}

void fatal_error(char *msg)
{
	fprintf(stderr, "fatal error: %s\n", msg);
	if (errno)
		perror("");
	exit(1);
}

void die(char *msg)
{
	printf("%s\n", msg);
	exit(0);
}

void printd(char *msg)
{
	fprintf(stderr, "%s\n", msg);
}



/* copyright (c) 2001  
 * marius aamodt eriksen
 * all rights reserved
 * 
 * permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any purpose,
 * so long as the name of the author is not used in any advertising or 
 * publicity pertaining to the use or distribution of this software 
 * without specific, written prior authorization.  if the above copyright 
 * notice or any other identification of the author is included in any 
 * copy of any portion of this software, then the disclaimer below must 
 * also be included.
 * 
 * this software is provided as is, without representation from the
 * the author as to its fitness for any purpose, and withoutwarranty by 
 * the author of any kind, either express or implied, including without 
 * limitation the implied warranties of merchantability and fitness for 
 * a particular purpose.  the author shall not be liable for any damages,
 * including special, indirect, incidental, or consequential damages,
 * with respect to any claim arising out or in connection with the use
 * of the software, even if it has been or is hereafter advised of the
 * possibility of such damages. */
